import { Utils } from "@appsensorlike/appsensorlike/utils/Utils.js";
import { Logger } from "@appsensorlike/appsensorlike/logging/logging.js";
import { ActionRequest, ActionResponse, UUID_QUERY_PARAM } from "../appsensor-websocket.js";
import WebSocket from "ws";
import { v4 as uuidv4 } from 'uuid';
class WebSocketClientConfig {
    constructor() {
        this.address = '';
        this.reconnectOnConnectionLost = true;
        this.reconnectRetryInterval = WebSocketClientConfig.DEFAULT_RETRY_INTERVAL;
    }
    checkValidInitialize() {
        if (this.reconnectOnConnectionLost === true) { //mind undefined value
            if (!this.reconnectRetryInterval) {
                this.reconnectRetryInterval = WebSocketClientConfig.DEFAULT_RETRY_INTERVAL;
            }
        }
    }
}
WebSocketClientConfig.DEFAULT_RETRY_INTERVAL = 20000;
class AppSensorWebSocketClient {
    constructor(config) {
        this.socket = null;
        this.reconnectTimer = null;
        this.myUUID = uuidv4();
        const _address = config.address + '?' + UUID_QUERY_PARAM + '=' + this.myUUID;
        this.reconnectAddress = _address;
        this.reconnectConfig = config;
        this.connect(_address, config.options);
    }
    connect(address, options) {
        this.socket = new WebSocket(address, options);
        const _myUUID = this.myUUID;
        this.socket.on('open', this.onOpen.bind(this));
        this.socket.on('error', this.onError.bind(this));
        this.socket.on('message', this.onMessage.bind(this));
        this.socket.on('close', this.onClose.bind(this));
    }
    onOpen() {
        if (this.reconnectTimer) {
            clearInterval(this.reconnectTimer);
            this.reconnectTimer = null;
            Logger.getClientLogger().info('AppSensorWebSocketClient.onOpen:', 'Reconnected');
        }
        Logger.getClientLogger().trace(`AppSensorWebSocketClient.socket: ${this.myUUID}:`, 'open');
    }
    onError(error) {
        Logger.getClientLogger().error(`AppSensorWebSocketClient.socket: ${this.myUUID}:`, 'error ', error);
    }
    onClose(code, reason) {
        Logger.getClientLogger().trace(`AppSensorWebSocketClient.socket: ${this.myUUID}:`, 'close', ' CODE: ', code, ' REASON: ', reason.toString());
        if (this.reconnectConfig && this.reconnectConfig.reconnectOnConnectionLost &&
            !this.reconnectTimer) {
            this.reconnectTimer = setInterval(this.reconnect.bind(this), this.reconnectConfig.reconnectRetryInterval);
        }
    }
    reconnect() {
        Logger.getClientLogger().info('AppSensorWebSocketClient.reconnect:', 'Retry reconnect...');
        return this.connect(this.reconnectAddress, this.reconnectConfig.options);
    }
    onMessage(data, isBinary) {
        Logger.getClientLogger().trace(`AppSensorWebSocketClient.socket: ${this.myUUID}:`, 'message');
        const response = JSON.parse(data.toString());
        Object.setPrototypeOf(response, ActionResponse.prototype);
        if (response.accessDenied) {
            Logger.getClientLogger().warn('Access denied for this client application! Configure server!');
        }
        else if (response.unauthorizedAction) {
            Logger.getClientLogger().warn(`This client application is not authorized to perform '${response.actionName}' on server! Configure server!`);
        }
        else {
            this.onServerResponse(response);
        }
    }
    onServerResponse(response) {
        //your code goes here
    }
    static createRequest(actionName, parameters) {
        const uuid = uuidv4();
        return new ActionRequest(uuid, actionName, parameters);
    }
    async sendRequest(request, cb) {
        let waited = 0;
        const timeout = 500;
        if (!this.socket) {
            return;
        }
        while (this.socket.readyState === WebSocket.CONNECTING && waited < 5000) {
            await Utils.sleep(timeout);
            waited += timeout;
        }
        if (this.socket.readyState === WebSocket.CONNECTING) {
            cb(new Error('WebSocket in CONNECTING state for too long!'));
        }
        else if (this.socket.readyState === WebSocket.CLOSING) {
            cb(new Error('WebSocket in CLOSING state!'));
        }
        else if (this.socket.readyState === WebSocket.CLOSED) {
            cb(new Error('WebSocket has already been closed!'));
        }
        else if (this.socket.readyState === WebSocket.OPEN) {
            this.socket.send(JSON.stringify(request), cb);
        }
    }
    async closeSocket() {
        if (this.socket) {
            this.socket.close();
        }
    }
}
export { AppSensorWebSocketClient, WebSocketClientConfig };
